using System;
using _3DEngine;

namespace PlanetSystem
{

	public class Calculation
	{

		private static long mulDivCount;
		private static long trigCount;
		private static bool isLogging;

		static Calculation() 
		{
			mulDivCount = 0;
			trigCount = 0;
			isLogging = false;
		}

		public static void BeginOperationLog() 
		{
			mulDivCount = 0;
			trigCount = 0;
			isLogging = true;
		}

		public static void EndOperationLog() 
		{
			isLogging = false;
		}

		public static long GetMulDivCount() 
		{
			return mulDivCount;
		}

		public static long GetTrigCount() 
		{
			return trigCount;
		}

		private const double DURATION = 400;
		private const double TIME_RESOLUTION = 0.25;
		private const double EARTH_INCLINATION = 10;
		private const double EARTH_RADIUS = 2;
		private const double EARTH_ROTATION = 3;
		private const double MOON_INCLINATION = 16;
		private const double MOON_RADIUS = 1;
		private const double MOON_ROTATION = 10;
		private const double MOON_DISTANCE = 6;
		private const double MOON_TRACK_ROTATION = 11;

		private Point earthObserver;
		private Point earthCenter;

		private Point moonTarget;
		private Point moonCenter;

		private CoordinateSystem earthRotation;
		private Matrix earthRotationStep;

		private CoordinateSystem moonRotation;
		private Matrix moonRotationStep;
		private CoordinateSystem moonTrackTranslation;
		private Matrix moonDistanceTranslation;
		private Matrix moonTrackRotationStep;

		private CoordinateSystem universe;

		public Calculation() 
		{

			// Erzeuge den Beobachter
			earthObserver = new Point(EARTH_RADIUS,0,0);
			earthCenter = new Point(0,0,0);

			// Erzeuge den beobachteten Punkt
			moonTarget = new Point(MOON_RADIUS,0,0);
			moonCenter = new Point(0,0,0);

			// Erzeuge die Erde
			earthRotation = new CoordinateSystem();
			earthRotation.AddTransformable(earthObserver);
			earthRotation.AddTransformable(earthCenter);
			CoordinateSystem cs1 = new CoordinateSystem();
			cs1.AddTransformable(earthRotation);
			cs1.ChangeTransformation(Matrix.CreateRotateZMatrix(DegreeToRadians(EARTH_INCLINATION)));
			earthRotationStep = Matrix.CreateRotateYMatrix(DegreeToRadians(EARTH_ROTATION*TIME_RESOLUTION));

			// Erzeuge den Mond
			moonRotation = new CoordinateSystem();
			moonRotation.AddTransformable(moonTarget);
			moonRotation.AddTransformable(moonCenter);
			CoordinateSystem cs2 = new CoordinateSystem();
			cs2.AddTransformable(moonRotation);
			cs2.ChangeTransformation(Matrix.CreateRotateZMatrix(DegreeToRadians(MOON_INCLINATION)));
			moonRotationStep = Matrix.CreateRotateYMatrix(DegreeToRadians(MOON_ROTATION*TIME_RESOLUTION));
			moonTrackTranslation = new CoordinateSystem();
			moonTrackTranslation.AddTransformable(cs2);
			moonDistanceTranslation = Matrix.CreateTranslationMatrix(MOON_DISTANCE,0,0);
			moonTrackRotationStep = Matrix.CreateRotateYMatrix(DegreeToRadians(MOON_TRACK_ROTATION*TIME_RESOLUTION));

			// Erzeuge das Universum
			universe = new CoordinateSystem();
			universe.AddTransformable(cs1);
			universe.AddTransformable(moonTrackTranslation);
		}

		public Result[] Calculate() {

			Result[] result = new Result[1 + (int)(DURATION/TIME_RESOLUTION)];

			Matrix currentEarthRotation = Matrix.CreateIdentityMatrix();
			Matrix currentMoonRotation = Matrix.CreateIdentityMatrix();
			Matrix currentMoonTrackRotation = Matrix.CreateIdentityMatrix();
			Matrix currentMoonTrackTranslation = moonDistanceTranslation;

			for(int i=0; i<=DURATION/TIME_RESOLUTION; i++) 
			{
				Point.BeginOperationLog();
				Matrix.BeginOperationLog();
				BeginOperationLog();

				earthRotation.ChangeTransformation(currentEarthRotation);
				moonRotation.ChangeTransformation(currentMoonRotation);
				moonTrackTranslation.ChangeTransformation(currentMoonTrackTranslation);
				Matrix m = Matrix.CreateIdentityMatrix();
				universe.Transform(m);

				double a = CalculateAngle(earthCenter,earthObserver,moonTarget);
				bool visible;
				if(a < 0.0) 
				{
					visible = false;
				}
				else 
				{
					double b = CalculateAngle(moonCenter,moonTarget,earthObserver);
					if(b < 0.0) 
					{
						visible = false;
					}
					else 
					{
						visible = true;
					}
				}

				currentEarthRotation  = currentEarthRotation.Mult(earthRotationStep);
				currentMoonRotation  = currentMoonRotation.Mult(moonRotationStep);
				currentMoonTrackRotation = currentMoonTrackRotation.Mult(moonTrackRotationStep);
				currentMoonTrackTranslation = currentMoonTrackRotation.Mult(moonDistanceTranslation);

				Point.EndOperationLog();
				Matrix.EndOperationLog();
				EndOperationLog();

				result[i] = new Result(a, visible, Matrix.GetAddSubCount() + Point.GetAddSubCount(), Matrix.GetMulDivCount() + Point.GetMulDivCount() + GetMulDivCount(), Matrix.GetTrigCount() + GetTrigCount(), Point.GetSqrtCount());

			}

			return result;

		}

		static double DegreeToRadians(double degree) 
		{
			return (degree*Math.PI)/180.0;
		}

		static double RadiansToDegree(double radians) 
		{
			return (radians*180.0)/Math.PI;
		}

		// Negativer Winkel bedeutet unter der Tangentialebene
		static double CalculateAngle(Point center, Point observer, Point target) 
		{
			if(isLogging) 
			{
				mulDivCount +=2;
				trigCount ++;
			}
			Point n = observer.Sub(center);
			Point v = target.Sub(observer);
			double a = n.Mult(v);
			double sign = Math.Sign(a);
			double b = n.GetLength() * v.GetLength();
			double c = Math.Asin(Math.Abs(a)/b);
			return sign * RadiansToDegree(c);
		}

	}

}
